// Copyright 2010 Google Inc. All Rights Reserved.

/**
 * @fileoverview A handler for incoming audio display and additional remote
 * mute functionality within the video feeds.
 *
 * @author martino@google.com (Martin Ohman)
 */
goog.provide('talk.media.c3d.AudioDisplayHandler');
goog.provide('talk.media.c3d.AudioDisplayHandler.Event');
goog.provide('talk.media.c3d.AudioDisplayHandler.EventType');
goog.provide('talk.media.c3d.AudioDisplayHandler.VolumeDisplay');

goog.require('goog.Disposable');
goog.require('goog.debug.Logger');
goog.require('goog.events.Event');
goog.require('talk.media.c3d.FeedHandler');
goog.require('talk.media.c3d.Image');
goog.require('talk.media.c3d.ImageOverlay');
goog.require('talk.media.c3d.O3dBundle');
goog.require('talk.media.c3d.helpers');


/**
 * Creates an audio display handler for a video feed.
 *
 * @param {talk.media.c3d.O3dBundle} o3dBundle The collection of global O3D
 *     handles.
 * @param {talk.media.c3d.BaseFeed} feed The object representing the feed.
 * @param {!goog.events.EventTarget} eventTarget The event target for this
 *     handler.
 * @constructor
 * @extends {talk.media.c3d.FeedHandler}
 */
talk.media.c3d.AudioDisplayHandler = function(o3dBundle, feed, eventTarget) {
  goog.base(this, o3dBundle, feed);

  /**
   * Whether the mouse was pressed over the mute icon.
   * @type {boolean}
   * @private
   */
  this.armed_ = false;

  /**
   * An array of images used to display the volume.
   * @type {!Array.<talk.media.c3d.Image>}
   * @private
   */
  this.volumeIcons_ = [];

  /**
   * Counter for loaded volume icons.
   * @type {number}
   * @private
   */
  this.numLoadedVolumeIcons_ = 0;

  /**
   * The event target used to dispatch events from this handler.
   * @type {!goog.events.EventTarget}
   * @private
   */
  this.eventTarget_ = eventTarget;
};
goog.inherits(talk.media.c3d.AudioDisplayHandler, talk.media.c3d.FeedHandler);


/**
 * Whether the remote mute experiment is enabled.
 *
 * @type {boolean}
 */
talk.media.c3d.AudioDisplayHandler.remoteMuteEnabled = false;


/**
 * Our logger.
 *
 * @type {goog.debug.Logger}
 * @private
 */
talk.media.c3d.AudioDisplayHandler.prototype.logger_ =
    goog.debug.Logger.getLogger('talk.media.c3d.AudioDisplayHandler');


/**
 * An image displaying muted state.
 *
 * @type {talk.media.c3d.Image}
 * @private
 */
talk.media.c3d.AudioDisplayHandler.prototype.muteIcon_;


/**
 * The volume display for this handler.
 *
 * @type {talk.media.c3d.AudioDisplayHandler.VolumeDisplay}
 * @private
 */
talk.media.c3d.AudioDisplayHandler.prototype.volumeDisplay_ = null;


/**
 * Initializes images and event handlers.
 * Adds volume display icons and a clickable mute icon to the video area.
 * @param {talk.media.c3d.ImageOverlay=} opt_imageOverlay The layer for images.
 */
talk.media.c3d.AudioDisplayHandler.prototype.init = function(opt_imageOverlay) {
  goog.base(this, 'init', opt_imageOverlay);

  this.volumeIcons_[0] = this.createImage_(this.imageOverlay_,
      'images/volume_meter0.png', 3, 3);
  this.volumeIcons_[1] = this.createImage_(this.imageOverlay_,
      'images/volume_meter1.png', 3, 3);
  this.volumeIcons_[2] = this.createImage_(this.imageOverlay_,
      'images/volume_meter2.png', 3, 3);
  this.volumeIcons_[3] = this.createImage_(this.imageOverlay_,
      'images/volume_meter3.png', 3, 3);
  this.volumeIcons_[4] = this.createImage_(this.imageOverlay_,
      'images/volume_meter4.png', 3, 3);
  this.volumeIcons_[5] = this.createImage_(this.imageOverlay_,
      'images/volume_meter5.png', 3, 3);

  this.muteIcon_ = this.createImage_(this.imageOverlay_, 'images/mute_mic.png',
      -3, 3);
};


/**
 * @override
 * @protected
 */
talk.media.c3d.AudioDisplayHandler.prototype.onMouseMove = function(e) {
  this.muteIcon_.setVisible(this.feed.hitTest(e['x'], e['y']));
};


/**
 * @override
 * @protected
 */
talk.media.c3d.AudioDisplayHandler.prototype.onMouseDown = function(e) {
  var pos = this.translateToFeedPoint_([e['x'], e['y'], 0]);
  if (this.muteIcon_.hitTest(pos[0], pos[1])) {
    this.armed_ = true;
  }
};


/**
 * @override
 * @protected
 */
talk.media.c3d.AudioDisplayHandler.prototype.onMouseUp = function(e) {
  if (this.armed_) {
    var pos = this.translateToFeedPoint_([e['x'], e['y'], 0]);
    if (this.muteIcon_.hitTest(pos[0], pos[1])) {
      this.eventTarget_.dispatchEvent(
          new talk.media.c3d.AudioDisplayHandler.Event(
              talk.media.c3d.AudioDisplayHandler.EventType.MUTE_AUDIO,
              this.feed.getEndpoint().getParticipant().getJid()));
    }
  }
  this.armed_ = false;
};


// TODO(martino): Investigate if methods from here on could be moved to
// FeedHandler, and do the same thing in PipHandler.
/**
 * Creates an icon based on the given image URL.
 *
 * @param {talk.media.c3d.ImageOverlay} imageOverlay The overlay to display the
 *     image.
 * @param {string} url The url of the image.
 * @param {number} x The initial x coordinate for this image.
 * @param {number} y The initial y coordinate for this image.
 * @return {talk.media.c3d.Image} A new Image object.
 * @private
 */
talk.media.c3d.AudioDisplayHandler.prototype.createImage_ = function(
    imageOverlay, url, x, y) {
  var image = imageOverlay.createImage(0, 0, url, goog.bind(function() {
    this.imageLoaded_(image, x, y);
  }, this));
  return image;
};


/**
 * Moves image to the given initial position. Called when an image has
 * completed loading.
 *
 * @param {talk.media.c3d.Image} image The image that loaded.
 * @param {number} x The new horizontal position.
 * @param {number} y The new vertical position.
 * @private
 */
talk.media.c3d.AudioDisplayHandler.prototype.imageLoaded_ = function(image,
    x, y) {
  var feedAttrs = this.feed.getFeedAttrs();
  image.resize(image.textureWidth(), image.textureHeight());
  image.moveImageWithEdgeWrapping(x, y, feedAttrs.w, feedAttrs.h);
  image.setVisible(false);

  if (image != this.muteIcon_) {
    if (++this.numLoadedVolumeIcons_ == this.volumeIcons_.length) {
      this.volumeDisplay_ =
          new talk.media.c3d.AudioDisplayHandler.VolumeDisplay(
              this.volumeIcons_);
      this.eventTarget_.dispatchEvent(
          new talk.media.c3d.AudioDisplayHandler.Event(
              talk.media.c3d.AudioDisplayHandler.EventType.VOLUME_DISPLAY_ADDED,
              this.feed.getEndpoint().getParticipant().getJid(),
              this.volumeDisplay_));
    }
  }
};


/**
 * Translates world vector to feed local.
 *
 * @param {Array.<number>} vector The world vector to be translated.
 * @return {Array.<number>} The translated vector.
 * @private
 */
talk.media.c3d.AudioDisplayHandler.prototype.translateToFeedPoint_ = function(
    vector) {
  var inverseMatrix = talk.media.c3d.inverse4(
      this.feed.feedObject.getTransform()['localMatrix']);
  return talk.media.c3d.transformPoint(inverseMatrix, vector);
};


/**
 * Removes all references to C3D objects.
 */
talk.media.c3d.AudioDisplayHandler.prototype.disposeC3d = function() {
  goog.base(this, 'disposeC3d');

  for (var i = 0; i < this.volumeIcons_.length; i++) {
    this.volumeIcons_[i].disposeC3d();
  }
  this.volumeIcons_ = [];

  if (this.muteIcon_) {
    this.muteIcon_.disposeC3d();
    this.muteIcon_ = null;
  }
};


/**
 * @override
 * @protected
 */
talk.media.c3d.AudioDisplayHandler.prototype.disposeInternal = function() {
  goog.base(this, 'disposeInternal');

  this.disposeC3d();

  if (this.volumeDisplay_) {
    this.volumeDisplay_.dispose();
    delete this.volumeDisplay_;
  }

  this.eventTarget_.dispatchEvent(
      new talk.media.c3d.AudioDisplayHandler.Event(
          talk.media.c3d.AudioDisplayHandler.EventType.VOLUME_DISPLAY_REMOVED,
          this.feed.getEndpoint().getParticipant().getJid()));
};


/**
 * Dispatched by the AudioDisplayHandler.
 *
 * @param {!talk.media.c3d.AudioDisplayHandler.EventType} type The type of this
 *     event.
 * @param {string} jid The jid of the feed where this event originates from.
 * @param {talk.media.c3d.AudioDisplayHandler.VolumeDisplay=} opt_volumeDisplay
 *     An optional control to display the volume of the feed of this event.
 * @constructor
 * @extends goog.events.Event
 */
talk.media.c3d.AudioDisplayHandler.Event = function(type, jid,
    opt_volumeDisplay) {
  goog.base(this, type);

  /**
   * The jid of the feed of the handler dispatching this event.
   * @type {string}
   */
  this.jid = jid;

  /**
   * The volume display control of the event dispatching handler.
   * @type {talk.media.c3d.AudioDisplayHandler.VolumeDisplay|undefined}
   */
  this.volumeDisplay = opt_volumeDisplay;
};
goog.inherits(talk.media.c3d.AudioDisplayHandler.Event, goog.events.Event);



/**
 * Constants for AudioDisplayHandler event types.
 * @enum {string}
 */
talk.media.c3d.AudioDisplayHandler.EventType = {
  VOLUME_DISPLAY_ADDED: 'volumeDisplayAdded',
  VOLUME_DISPLAY_REMOVED: 'volumeDisplayRemoved',
  MUTE_AUDIO: 'muteAudio'
};



/**
 * Creates a wrapper to control the volume display.
 *
 * @param {!Array.<talk.media.c3d.Image>} icons An array of volume icons.
 * @constructor
 * @extends {goog.Disposable}
 */
talk.media.c3d.AudioDisplayHandler.VolumeDisplay = function(icons) {
  /**
   * @type {!Array.<talk.media.c3d.Image>}
   * @private
   */
  this.volumeIcons_ = icons;

  /**
   * @type {talk.media.c3d.Image}
   * @private
   */
  this.shownIcon_ = this.volumeIcons_[0];
  this.shownIcon_.setVisible(true);
};
goog.inherits(talk.media.c3d.AudioDisplayHandler.VolumeDisplay,
    goog.Disposable);

/**
 * Shows or hides the appropriate icon.
 * @param {number} volume An integer between 0 (silent) and 5.
 */
talk.media.c3d.AudioDisplayHandler.VolumeDisplay.prototype.setVolume =
    function(volume) {
  if (this.shownIcon_ != this.volumeIcons_[volume]) {
    this.shownIcon_.setVisible(false);
    this.shownIcon_ = this.volumeIcons_[volume];
    this.shownIcon_.setVisible(true);
  }
};


/**
 * @override
 * @protected
 */
talk.media.c3d.AudioDisplayHandler.VolumeDisplay.prototype.disposeInternal =
    function() {
  goog.base(this, 'disposeInternal');
  delete this.volumeIcons_;
};


