// Copyright 2007 Google Inc.
// All Rights Reserved.

/**
 * @fileoverview A Flash UI object handling real time communication.
 *
 * @author jessan@google.com (Jessan Hutchison-Quillian)
 */

goog.provide('talk.media.CommunicationPlayer');

goog.require('goog.debug.Logger');
goog.require('goog.debug.Trace');
goog.require('goog.events.EventHandler');
goog.require('goog.json');
goog.require('talk.media.CallManager');
goog.require('talk.media.CallManager.InstallNeededEvent');
goog.require('talk.media.CallType');
goog.require('talk.media.FlashObject.Wmodes');
goog.require('talk.media.FlashPlayer');
goog.require('talk.media.Participants');
goog.require('talk.media.UiMode');

/**
 * Creates a player that displays real time communication (voice/video calls).
 *
 * @param {talk.media.CallManager} callManager The manager for voice/video
 *     calls.
 * @param {string} opt_flashVars Parameters to be passed to the SWF.
 *     The string should be in URL parameter format: var1=A&var2=B...
 * @param {string} opt_swfUrl The URL that points to the media-player.swf.
 * @constructor
 * @extends {talk.media.FlashPlayer}
 */
talk.media.CommunicationPlayer = function(callManager, opt_flashVars,
                                          opt_swfUrl) {


  /**
   * The location of our media player.  It varies based on the container (ie
   * Gmail vs Supermoles).
   * @type {string}
   */
  this.swfUrl_ = (opt_swfUrl ? opt_swfUrl : 'im/media-player.swf') + '?ver=' +
      talk.media.CommunicationPlayer.CALL_PLAYER_VERSION_;

  /**
   * A trace of how long it takes us to start video from the time that we
   * are created
   * @type number?
   */
  this.videoStartTrace_ = goog.debug.Trace.startTracer(
      'communicationPlayer.start');

  talk.media.FlashPlayer.call(this, this.getName(), this.getSwfUrl(),
      'plugin=true' + (opt_flashVars ? '&' + opt_flashVars : ''));

  /**
   * The call manager handles communication with the server and the
   * TalkPlugin.
   *
   * @type talk.media.CallManager
   * @private
   */
  this.callManager_ = callManager;


  /**
   * The UI mode for this player.
   *
   * @type {talk.media.UiMode}
   * @private
   */
  this.uiMode_ = talk.media.UiMode.NORMAL;


  /**
   * EventHandler to simplify installing/removing event listeners.
   * @type goog.events.EventHandler
   * @private
   */
  this.eventHandler_ = new goog.events.EventHandler(this);

  this.eventHandler_.listen(this.callManager_,
      talk.media.CallManager.Event.SHOW_VIDEO, this.handleShowVideo_);

  this.sendToPlayer(talk.media.CommunicationPlayer.SendMsg.PLUGIN_DATA,
      this.callManager_.getFluteVersionInfo());

};
goog.inherits(talk.media.CommunicationPlayer, talk.media.FlashPlayer);

/**
 * A bogus "version" for the SWF, used in a URL param to download it. The sole
 * purpose of this is to break browser caching when we upgrade the SWF. The URL
 * param is not actually used for anything.
 * TIP(tschmelcher): Set this to goog.now() when testing SWF changes to force
 * it to be reloaded in each new Gmail session.
 *
 * @type {string}
 * @const
 * @private
 */
talk.media.CommunicationPlayer.CALL_PLAYER_VERSION_ = '2.0.5';

talk.media.CommunicationPlayer.prototype.logger_ =
    goog.debug.Logger.getLogger('talk.media.CommunicationPlayer');

/**
 *
 * @enum {string}
 */
talk.media.CommunicationPlayer.SendMsg = {

  /**
   * A message to begin displaying call UI.
   */
  START_CALL : 'SC',

  /**
   * A message indicating the player should update its UI in response to user
   * action.
   */
  UI_COMMAND : 'UC',

  /**
   * A message to show an avatar.
   */
  SHOW_AVATAR : 'SA',

  /**
   * A message to update the participants.
   */
  UPDATE_PARTICIPANTS : 'UP',

  /**
   * A message containing data from the server (jingle messages).
   */
  JINGLE : 'connect-data',

  /**
   * Messages that come from the Google Talk plugin.
   */
  PLUGIN_DATA : 'plugin-data',

  /**
   * Microphone mute state message for responding to state requests.
   */
  MUTE_STATE : 'MUTE_STATE'
};


/**
 *
 * @enum {string}
 */
talk.media.CommunicationPlayer.RecvMsg = {
  /**
   * A message containing data for the server (jingle messages)
   */
  JINGLE : 'connect-data',

  /**
   * Messages should be sent to the Google Talk plugin.
   */
  PLUGIN_DATA : 'plugin-data',

  /**
   * A message containing data for the hosting chrome (us), that need not be
   * forwarded to the server or plugin.
   */
  SESSION_DATA : 'session-data',

  FEED_DISPLAYED : 'FEED',

  /**
   * Either get or set of microphone mute state, indicated by 'g' or 's' as
   * first item of arguments.
   */
  MUTE_STATE : 'MUTE_STATE'
};


/**
 * Events generated by the CommunicationPlayer.
 *
 * @enum {string}
 */
talk.media.CommunicationPlayer.Events = {
  /**
   * Message that player is being disposed.
   */
  DISPOSED : 'DP'
};


/**
 *
 * @enum {string}
 * @private
 */
talk.media.CommunicationPlayer.SessionMessage_ = {
  /**
   * A message indicating the user attempted to enter full screen with an
   * incompatible Flash version and we should prompt them to upgrade.
   */
  FULLSCREEN_UPGRADE : 'fs'
};


/**
 * The ID of the call being displayed.
 *
 * @type string
 * @private
 */
talk.media.CommunicationPlayer.prototype.sessionId_;


/**
 * The JID of the user we're calling.
 *
 * @type string
 * @private
 */
talk.media.CommunicationPlayer.prototype.senderJid_;


/**
 * True iff this is a test call. Test calls were added after we solidified the
 * Flash codebase, so they use a different code path where we ignore a bunch of
 * Flash's messaging and instead do it ourselves in JavaScript.
 *
 * @type boolean
 * @private
 */
talk.media.CommunicationPlayer.prototype.isTestCall_;


/**
 * @type {talk.media.FlashObject.Wmodes}
 * @private
 */
talk.media.CommunicationPlayer.wmode_ = talk.media.FlashObject.Wmodes.WINDOW;


/**
 * Sets the wmode parameter that will be used when instantiating the player SWF.
 *
 * @param {talk.media.FlashObject.Wmodes} wmode
 */
talk.media.CommunicationPlayer.setWmode = function(wmode) {
  talk.media.CommunicationPlayer.wmode_ = wmode;
};


/**
 * Returns the URL to the flute-enabled media player.
 *
 * @return {string} The version specific URL.
 */
talk.media.CommunicationPlayer.prototype.getSwfUrl = function() {
  return this.swfUrl_;
};


/**
 * Override.
 */
talk.media.CommunicationPlayer.prototype.getName = function() {
  return 'yj_c_player';
};


/**
 * Returns the wmode parameter to use when writing the Flash DOM object.
 *
 * @return {talk.media.FlashObject.Wmodes} The Wmode to use.
 */
talk.media.CommunicationPlayer.prototype.getWmode = function() {
  return talk.media.CommunicationPlayer.wmode_;
};


/**
 * Returns the UI mode for this player.
 *
 * @return {talk.media.UiMode} The UI mode for this player.
 */
talk.media.CommunicationPlayer.prototype.getUiMode = function() {
  return this.uiMode_;
};


/**
 * Sets the UI mode for this player.
 *
 * @param {talk.media.UiMode} uiMode The UI mode for this player.
 */
talk.media.CommunicationPlayer.prototype.setUiMode = function(uiMode) {
  this.uiMode_ = uiMode;
};


/**
 * Handles a new call.
 *
 * @param {boolean} isOutgoing If the call is placed by the local user or not.
 * @param {string} sessionId The id of the call session.
 * @param {string} senderJid The JID being called.
 * @param {talk.media.Participants} participants The users in the call.
 * @param {talk.media.CallType} callType The type of call.
 * @param {boolean} isMultiUser Whether the call is a MUC call.
 */
talk.media.CommunicationPlayer.prototype.onNewCall = function(isOutgoing,
    sessionId, senderJid, participants, callType, isMultiUser) {

  this.sessionId_ = sessionId;
  this.senderJid_ = senderJid;
  this.isTestCall_ = (callType == talk.media.CallType.TEST);
  var participantArray = this.getParticipantArray_(participants);
  this.sendToPlayer(talk.media.CommunicationPlayer.SendMsg.START_CALL,
      [isOutgoing, sessionId, participantArray,
      this.isTestCall_ ? talk.media.CallType.VIDEO : callType, isMultiUser]);

  if (this.isTestCall_) {
    // TODO(tschmelcher): CallManager should control test calls.
    this.callManager_.sendFluteMsg(goog.json.serialize(
        ['jn', this.senderJid_, this.sessionId_, 't', 's']));
  }

  if (callType != talk.media.CallType.VIDEO &&
      callType != talk.media.CallType.TEST) {
    this.stopTracer_();
  }
};


/**
 * Begins displaying the already active call.
 *
 * @param {talk.media.Participants} participants The users in the active call.
 * @param {boolean} isMultiUser Whether the call is a MUC call.
 */
talk.media.CommunicationPlayer.prototype.displayActiveCall = function(
    participants, isMultiUser) {
  var activeCall = this.callManager_.getActiveCall();
  if (activeCall) {
    // Reset the SWF state so that it will re-init correctly and not miss the
    // START_CALL message.
    this.resetState();
    this.sendToPlayer(
        talk.media.CommunicationPlayer.SendMsg.PLUGIN_DATA,
        this.callManager_.getFluteVersionInfo());
    this.sessionId_ = activeCall.sessionId;
    this.senderJid_ = activeCall.remoteJids[0];
    this.isTestCall_ = (activeCall.type == talk.media.CallType.TEST);
    var participantArray = this.getParticipantArray_(participants);
    this.sendToPlayer(talk.media.CommunicationPlayer.SendMsg.START_CALL, [false,
      activeCall.sessionId, participantArray,
      this.isTestCall_ ? talk.media.CallType.VIDEO : activeCall.type,
      isMultiUser, activeCall.state]);
    if (activeCall.type != talk.media.CallType.VIDEO &&
        activeCall.type != talk.media.CallType.TEST) {
      this.stopTracer_();
    }
  } else {
    this.logger_.warning('displayActiveCall called with no active call');
  }
};


/**
 * Turns on a user's video stream in the Flute.
 *
 * @param {string} ssrc The SSRC of the video stream.
 * @private
 */
talk.media.CommunicationPlayer.prototype.fluteVideoStreamOn_ = function(ssrc) {
  this.callManager_.sendFluteMsg(goog.json.serialize([
      'streamon', this.senderJid_, this.sessionId_, 1, ssrc]));
};


/**
 * Turns off a user's video stream in the Flute.
 *
 * @param {string} ssrc The SSRC of the video stream.
 * @private
 */
talk.media.CommunicationPlayer.prototype.fluteVideoStreamOff_ = function(ssrc) {
  this.callManager_.sendFluteMsg(goog.json.serialize([
      'streamoff', this.senderJid_, this.sessionId_, 1, ssrc]));
};


/**
 * Handles a request to show a video stream.
 *
 * @param {talk.media.CallManager.ShowVideoEvent} e The event.
 * @private
 */
talk.media.CommunicationPlayer.prototype.handleShowVideo_ = function(e) {
  if (e.sessionId == this.sessionId_) {
    this.fluteVideoStreamOn_(e.ssrc);
  }
};


/**
 * Shows the avatar for a given jid.
 *
 * @param {string} jid The jid whose avatar to show.
 */
talk.media.CommunicationPlayer.prototype.showAvatar = function(jid) {
  this.sendToPlayer(talk.media.CommunicationPlayer.SendMsg.SHOW_AVATAR, [jid]);
};


/**
 * Updates the list of participants.
 *
 * @param {talk.media.Participants} participants The list of participants.
 */
talk.media.CommunicationPlayer.prototype.updateParticipants = function(
    participants) {
  var participantArray = this.getParticipantArray_(participants);
  this.sendToPlayer(talk.media.CommunicationPlayer.SendMsg.UPDATE_PARTICIPANTS,
      [participantArray]);
};


/**
 * Relays a jingle signaling or talk plugin message to the SWF.
 *
 * @param {string} type The kind of msg.
 * @param {string} message The json message to send to the SWF.
 */
talk.media.CommunicationPlayer.prototype.relayCallMessage = function(type,
                                                                     message) {
  this.sendToPlayer(type, goog.json.parse(/** @type {Object} */(message)));
};


/**
 * @override
 */
talk.media.CommunicationPlayer.prototype.handleMessage = function(type, msg) {
  switch (type) {
    case talk.media.CommunicationPlayer.RecvMsg.JINGLE:
      if (!this.isTestCall_ &&
          // Historically Flash's job, but not anymore
          msg[1][0] != 'jn' && msg[1][0] != 'ja') {
        this.callManager_.sendJingleMsg(goog.json.serialize(
            /** @type {Object} */(msg[1])));
      }
      break;
    case talk.media.CommunicationPlayer.RecvMsg.PLUGIN_DATA:
      // For test calls, the only message that we let Flash send is the streamon
      // for the local feed. The "0" here is its magiccam number in the JSON
      // message.
      if ((!this.isTestCall_ || (msg[1][0] == 'streamon' && msg[1][3] == 0)) &&
          // Historically Flash's job, but not anymore
          msg[1][0] != 'jn' && msg[1][0] != 'ja') {
        this.callManager_.sendFluteMsg(goog.json.serialize(
            /** @type {Object} */(msg[1])));
      }
      break;
    case talk.media.CommunicationPlayer.RecvMsg.SESSION_DATA:
      // Do we want to handle these messages here?  In the callmanager instead?
      if (!this.isTestCall_ && msg[1] ==
          talk.media.CommunicationPlayer.SessionMessage_.FULLSCREEN_UPGRADE) {
        var call = this.callManager_.getActiveCall();
        var e = new talk.media.CallManager.InstallNeededEvent(
            talk.media.CallManager.InstallNeededType.FULLSCREEN_UPGRADE_NEEDED,
            false, call.remoteJids[0], call.type == talk.media.CallType.VIDEO);

        this.callManager_.dispatchEvent(e);
      }
      break;
    case talk.media.CommunicationPlayer.RecvMsg.FEED_DISPLAYED:
      this.logger_.info('feed displayed message from Flash');
      this.stopTracer_();
      break;
    case talk.media.CommunicationPlayer.RecvMsg.MUTE_STATE:
      if (msg[1] == 's') {
        this.logger_.info('set mute state from Flash (muted = ' + msg[2] + ')');
        this.callManager_.setMicrophoneMute(!!msg[2]);
      } else {
        var muted = this.callManager_.getMicrophoneMute();
        this.logger_.info('request for mute state from Flash (muted = ' +
            muted + ')');
        this.sendToPlayer(talk.media.CommunicationPlayer.SendMsg.MUTE_STATE,
            ['s', muted]);
      }
      break;
    default:
      return false;
  }
  return true;
};


talk.media.CommunicationPlayer.prototype.stopTracer_ = function() {
  if (this.videoStartTrace_ != null) {
    goog.debug.Trace.stopTracer(this.videoStartTrace_, 0);
    delete this.videoStartTrace_;
    this.logger_.info(goog.debug.Trace.getFormattedTrace());
  }
};


/**
 * Converts the participants to an array formatted for Flash.
 *
 * @param {talk.media.Participants} participants The participants in the call.
 * @return {Array.<Array>} The users in the call. Index 0 is the local user.
 *     The elements are [{string} JID, {boolean} isLocal, {string} avatarUrl].
 * @private
 */
talk.media.CommunicationPlayer.prototype.getParticipantArray_ =
    function(participants) {
  var ids = participants.getKeys();
  var participantArray = [];
  for (var i = 0; i < ids.length; ++i) {
    var participant = participants.get(ids[i]);
    if (participant.isLocal()) {
      participantArray.push(
          [participant.getJid(), true, participant.getAvatar()]);
      break;
    }
  }
  for (var i = 0; i < ids.length; ++i) {
    var participant = participants.get(ids[i]);
    if (!participant.isLocal()) {
      participantArray.push(
          [participant.getJid(), false, participant.getAvatar()]);
      break;
    }
  }
  return participantArray;
};


/**
 * @override
 */
talk.media.CommunicationPlayer.prototype.disposeInternal = function() {
  this.dispatchEvent(talk.media.CommunicationPlayer.Events.DISPOSED);
  talk.media.CommunicationPlayer.superClass_.disposeInternal.call(this);
  this.eventHandler_.dispose()
  this.eventHandler_ = null;
  this.callManager_ = null;
};
