// Copyright 2009 Google Inc.
// All Rights Reserved.

/**
 * @fileoverview Contains information about the participants in a call/meeting.
 *
 * @author mikaeld@google.com (Mikael Drugge)
 * @author geer@google.com (Arthur van der Geer)
 */

goog.provide('talk.media.Participants');

goog.require('goog.debug.Logger');
goog.require('goog.structs.Map');
goog.require('talk.Jid');
goog.require('talk.media.Participant');


/**
 * The Participants.
 *
 * @constructor
 */
talk.media.Participants = function() {
  /**
   * The structure that contains the participants.
   *
   * @type {!goog.structs.Map}
   * @private
   */
  this.map_ = new goog.structs.Map();
};


/**
 * Logger for debugging.
 *
 * @type {!goog.debug.Logger}
 * @private
 */
talk.media.Participants.prototype.logger_ =
    goog.debug.Logger.getLogger('talk.media.Participants');


/**
 * Returns the given participant.
 *
 * @param {string} id The participant id.
 * @return {talk.media.Participant} The participant, or null if not found.
 */
talk.media.Participants.prototype.get = function(id) {
  if (!this.map_.containsKey(id)) {
    return null;
  }
  return /** @type {talk.media.Participant} */ (this.map_.get(id));
};


/**
 * Adds a participant. If a participant with the same id has already been
 * added, the previous information will be overwritten by the new information.
 *
 * @param {string} id The participant id.
 * @param {talk.media.Participant} participant The participant.
 */
talk.media.Participants.prototype.add = function(id, participant) {
  this.map_.set(id, participant);
  this.logger_.info('Added participant: ' + id +
      ', current count: ' + this.map_.getCount());
};


/**
 * Removes a participant.
 *
 * @param {string} id The participant id.
 */
talk.media.Participants.prototype.remove = function(id) {
  this.map_.remove(id);
  this.logger_.info('Removed participant: ' + id +
        ', remaining count: ' + this.map_.getCount());
};


/**
 * Removes all participants.
 */
talk.media.Participants.prototype.removeAll = function() {
  var count = this.map_.getCount();
  this.map_.clear();
  this.logger_.info('Removed all ' + count + ' participants.');
};


/**
 * Returns the number of participants.
 *
 * @return {number} The number of participants.
 */
talk.media.Participants.prototype.getCount = function() {
  return this.map_.getCount();
};


/**
 * Returns the keys of all participant objects.
 *
 * @return {!Array.<string>} The key strings.
 */
talk.media.Participants.prototype.getKeys = function() {
  return this.map_.getKeys();
};


/**
 * Lookup the participants for a given jid.
 *
 * @param {string} jid The participant jid.
 * @return {!Array.<string>} The key strings.
 */
talk.media.Participants.prototype.findJid = function(jid) {
  var result = [];
  var ids = this.map_.getKeys();
  for (var i = 0; i < ids.length; ++i) {
    var participant =
        /** @type {talk.media.Participant} */ (this.map_.get(ids[i]));
    if (participant.getJid() == jid) {
      result.push(ids[i]);
    }
  }
  return result;
};


/**
 * Lookup the local participant.
 *
 * @return {talk.media.Participant} The local participant, or null if not found.
 */
talk.media.Participants.prototype.findLocal = function() {
  var ids = this.getKeys();
  for (var i = 0; i < ids.length; ++i) {
    var participant = this.get(ids[i]);
    if (participant.isLocal()) {
      return participant;
    }
  }
  return null;
};
